using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using NLog;
using System.IO;
using System.Collections;
using System.Net.Sockets;

namespace ChatSystem
{
    /// <summary>
    /// swb Modified !!!
    /// </summary>
    [Serializable]
    public class ChatUserItem
    {
        #region Constants

        public static string CLASS_NAME = typeof(ChatUserItem).ToString();

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Public Access

        #region Public Properties

        public string IdTagValue
        {
            get { return mIdTag; }
            set { mIdTag = value; }
        }

        public TcpClient ClientValue
        {
            get { return mClient; }
            set { mClient = value; }
        }

        public bool IsValidValue
        {
            get { return mIsValid; }
            set { mIsValid = value; }
        }


        #endregion

        #region Constructors

        /// <summary>
        /// Set default values for all fields.
        /// </summary>
        public ChatUserItem()
        {
            Reset();
        }

        /// <summary>
        /// Set all fields according to arguments
        /// </summary>
        /// <param name=""></param>
        public ChatUserItem(string aIdTag, TcpClient aClient, bool aIsValid)
        {
            Set(aIdTag, aClient, aIsValid);
        }

        /// <summary>
        /// Copy Constructor
        /// </summary>
        /// <param name=""></param>
        public ChatUserItem(ChatUserItem aItem)
        {
            Set(aItem);
        }

        #endregion

        #region Reset, Set

        /// <summary>
        /// Set fields to default values
        /// </summary>
        public void Reset()
        {
            mIdTag = Environment.MachineName+"-"+ Environment.UserName;
            mClient = new TcpClient();
            mIsValid = false;
        }

        /// <summary>
        /// Set the values of every field according to arguments
        /// </summary>
        /// <param name=""></param>
        public void Set(string aIdTag, TcpClient aClient, bool aIsValid)
        {
            mIdTag = aIdTag;
            mClient = aClient;
            mIsValid = aIsValid;
        }

        /// <summary>
        /// Copy our values from given ChatUserItem
        /// </summary>
        /// <param name="aFrom"></param>
        public void Set(ChatUserItem aFrom)
        {
            mIdTag = aFrom.mIdTag;
            mClient = aFrom.mClient;
            mIsValid = aFrom.mIsValid;
        }

        #endregion

        #region Logging

        /// <summary>
        /// Return ToString for this class, or "null"
        /// </summary>
        /// <param name="aValue"></param>
        /// <returns></returns>
        public static string ToStringP(ChatUserItem aValue)
        {
            if (aValue == null)
                return "null";
            else
                return aValue.ToString();
        }

        /// <summary>
        /// Return fields of this object as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "IdTag='"+mIdTag+"'"+
                ", mClient='"+mClient.ToString()+"', mIsValid="+mIsValid;
        }

        /// <summary>
        /// Log this object to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void Logit(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log this object to nlog - short version
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void LogitL(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitLToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log the contents of this class to String
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            sb.AppendLine(aIndent + aHeader);
            sb.AppendLine(aIndent + "    ChatUserItem:");
            sb.AppendLine(aIndent+"        "+"mIdTag='"+mIdTag+"'");
            sb.AppendLine(aIndent+"        "+"mClient='"+mClient.ToString()+"', mIsValid="+mIsValid);

            sb.Append(aIndent + "    End of ChatUserItem");
            return sb.ToString();
        }

        /// <summary>
        /// Log the contents of this class to String - short version
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitLToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            string msg;

            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            msg = "IdTag='" + mIdTag + "', IsValid=" + mIsValid;

            if (aHeader.Length > 0)
                sb.AppendLine(aIndent + aHeader + " " + msg);
            else
                sb.AppendLine(aIndent + msg);

            return sb.ToString();
        }

        #endregion

        #region I/O Routines

        /// <summary>
        /// Convert our class to an XML string via Serialize. Can throw exceptions
        /// </summary>
        /// <returns></returns>
        public string ToXML()
        {
            try
            {
                // Force UTF-8
                XmlSerializer serializer = new XmlSerializer(typeof(ChatUserItem));
                using (StringWriter writer = new IdLibrary.Utf8StringWriter())
                {
                    serializer.Serialize(writer, this);
                    return writer.ToString();
                }
            }
            catch (Exception ex)
            {
                logger.Error("ToXML: Exception '" + ex + "'");
                throw new Exception("ToXML", ex);
            }
        }

        /// <summary>
        /// Convert given xml string to our class. Throw exception on fail.
        /// </summary>
        /// <param name="xml"></param>
        /// <returns></returns>
        public ChatUserItem FromXML(string xml)
        {
            ChatUserItem ei = null;
            if ((xml != null) && (xml.Length > 0))
            {
                try
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(ChatUserItem));
                    ei = (ChatUserItem)serializer.Deserialize(new StringReader(xml));
                }
                catch (Exception ex)
                {
                    logger.Error("FromXML: Got exception '" + ex + "'");
                    throw new Exception("FromXML", ex);
                }
            }
            return ei;
        }

        /// <summary>
        /// Write out our class to given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool WriteFile(string fileName)
        {
            try
            {
                StreamWriter writer = new StreamWriter(fileName);
                writer.Write(this.ToXML());
                writer.Close();
            }
            catch (Exception ex)
            {
                logger.Info("WriteFile: Got exception '" + ex + "'");
                return false;
            }
            // Success
            return true;
        }

        /// <summary>
        /// Read our class from given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool ReadFile(string fileName)
        {
            try
            {
                StreamReader reader = new StreamReader(fileName);
                string contents = reader.ReadToEnd();
                reader.Close();

                // Set all our values from what we got back
                ChatUserItem fi = FromXML(contents);
                this.Set(fi);
            }
            catch (Exception ex)
            {
                logger.Info("ReadFile: Got exception '" + ex + "'");
                return false;
            }

            // Success
            return true;
        }

        #endregion

        #region Tester

        /// <summary>
        /// Basic test routine for ChatUserItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool Tester()
        {
            string TEST_FILE = "testChatUser.xml";
            bool result;
            ChatUserItem fi1, fi2, fi3;

            logger.Info("");
            logger.Info("Tester:1:");

            // Test constructors
            try
            {
                fi1 = new ChatUserItem();
                fi2 = BuildSample(1);
                fi3 = new ChatUserItem(fi2);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Constructor test got ex: " + ex);
                return false;
            }

            // Test logging
            try
            {
                logger.Info("Tester: ToStringP(fi2)=[" + ToStringP(fi2) + "]");
                logger.Info("");
                fi2.Logit(LogLevel.Info, "", "Tester: Logit without indent:");
                fi3.LogitL(LogLevel.Info, "   ", "Tester: LogitL with indent:");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Logit test got ex: " + ex);
                return false;
            }

            // Test Write and read
            try
            {
                logger.Info("");

                // Write out to XML
                if (fi2.WriteFile(TEST_FILE) == false)
                {
                    logger.Info("Tester: WriteFile(" + TEST_FILE + ") failed, returning false");
                    return false;
                }

                // Read in as different object
                result = fi1.ReadFile(TEST_FILE);

                if (result != true)
                {
                    logger.Info("Tester: Read " + TEST_FILE + " failed, returning false");
                    return false;
                }

                fi1.Logit(LogLevel.Info, "", "Tester: Read " + TEST_FILE + ":");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Read/Write test got ex: " + ex);
                return false;
            }

            // Success
            logger.Info("Tester: tests completed successfully");
            return true;
        }

        /// <summary>
        /// Return a sample record ChatUserItem.
        /// Return sample on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static ChatUserItem BuildSample(int aValue)
        {
            logger.Info("BuildSample: aValue='"+aValue+"'");
            string tempIdTag;
            TcpClient tempClient;

            ChatUserItem rec;

            if (aValue == 1)
            {
                tempIdTag = "aaa";
                tempClient = new TcpClient();

                rec = new ChatUserItem(tempIdTag, tempClient, false);
            }
            else if (aValue == 2)
            {
                tempIdTag = "bbb";
                tempClient = new TcpClient();

                rec = new ChatUserItem(tempIdTag, tempClient, false);
            }
            else if (aValue == 3)
            {
                tempIdTag = "ccc";
                tempClient = new TcpClient();

                rec = new ChatUserItem(tempIdTag, tempClient, false);
            }
            else
            {
                rec = new ChatUserItem("unused", new TcpClient(), false);
            }

            return rec;
        }

        /// <summary>
        /// Drill down through sample record until non-class 2nd item and update it
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public bool UpdateSample()
        {
            logger.Info("UpdateSample:1: mIdTag = bbb");
            mIdTag = "bbb";
            return true;
        }

        #endregion

        #region Class-Specific

        /// <summary>
        /// Get the input off the socket and return it as a message buffer.
        /// Return null on fail (IsValid set to false too)
        /// </summary>
        public MessageBuffer ReadStream()
        {
            byte[] buffer;
            NetworkStream ns;

            try
            {
                ns = mClient.GetStream();
                int numBytes = mClient.Available;
                buffer = new byte[numBytes];

                logger.Info("ReadStream:1: " + numBytes + " bytes available");

                // Read the message
                int received = ns.Read(buffer, 0, numBytes);

                if (received < 1)
                {
                    logger.Error("ReadStream: Got empty read, dropping user " + IdTagValue);
                    mIsValid = false;
                    return null;
                }

                // Load a message buffer that is easier to work with
                MessageBuffer.LogByteArray(buffer, LogLevel.Info, "    ", "Bytes read in:");
                MessageBuffer mb = new MessageBuffer();
                if (mb.LoadFromByteArray(buffer) == false)
                {
                    logger.Error("ReadStream: Failed to create message buffer, dropping user " + IdTagValue);
                    mIsValid = false;
                    return null;
                }

                mb.LogitL(LogLevel.Info, "", "ReadStream: MsgBuf after read:");
                return mb;
            }
            catch (Exception ex)
            {
                logger.Error("ReadStream:" + IdTagValue + ": Got ex: " + ex);
                mIsValid = false;
                return null;
            }
        }

        /// <summary>
        /// Write out this message to this user's socket. On success, return true.
        /// On error, sets IsValid to false and returns false.
        /// </summary>
        /// <param name="aBuf"></param>
        /// <returns></returns>
        public bool WriteStream(MessageBuffer aBuf)
        {
            NetworkStream ns;

            try
            {
                logger.Info("WriteStream:1: User " + IdTagValue + " sending msg [" + aBuf.ToString() + "]");
                ns = mClient.GetStream();

                if (ns == null)
                {
                    logger.Error("WriteStream:1: User " + IdTagValue + " cannot get stream to server");
                    return false;
                }

                byte[] outBytes = aBuf.SaveToByteArray();
                MessageBuffer.LogByteArray(outBytes, LogLevel.Info, "    ", "Bytes about to be written:");

                ns.Write(outBytes, 0, outBytes.Length);
                ns.Flush();

                return true;
            }
            catch (Exception ex)
            {
                logger.Error("WriteStream:" + IdTagValue + ": Got ex: " + ex);
                mIsValid = false;
                return false;
            }
        }

        /// <summary>
        /// Return our hostname-Username tag
        /// </summary>
        /// <returns></returns>
        //public string GetIdTag()
        //{
        //    return mHostName + "-" + mUserName;
        //}

        #endregion

        #endregion

        #region Private Access

        #region Private Members

        private string mIdTag;
        private TcpClient mClient;

        // Is this a valid entry
        private bool mIsValid;

        #endregion

        #endregion
    }
}
