using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using NLog;
using System.IO;
using System.Collections;
using System.Net.Sockets;

namespace ChatSystem
{
    /// <summary>
    /// swb modified!!
    /// </summary>
    [Serializable]
    public class ChatUserList
    {
        #region Constants

        public static string CLASS_NAME = typeof(ChatUserList).ToString();

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Public Access

        #region Public Properties

        public List<ChatUserItem> TheListValue
        {
            get { return mTheList; }
            set { mTheList = value; }
        }

        public int CurIdxValue
        {
            get { return mCurIdx; }
            set { mCurIdx = value; }
        }

        public string HeaderValue
        {
            get { return mHeader; }
            set { mHeader = value; }
        }

        [XmlIgnore]
        // Indexer to give array access to the main list.
        public ChatUserItem this[int index]
        {
            // Return the item at this index
            get
            {
                if (index < 0 || index >= mTheList.Count)
                    return null;

                return mTheList[index];
            }
            // Set the item at this index to given value
            set
            {
                if (index < 0 || index >= mTheList.Count)
                    return;

                mTheList[index] = value;
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Set default values for all fields.
        /// </summary>
        public ChatUserList()
        {
            Reset();
        }

        /// <summary>
        /// Set all fields according to arguments
        /// </summary>
        /// <param name=""></param>
        public ChatUserList(List<ChatUserItem> aTheList, int aCurIdx, string aHeader)
        {
            Set(aTheList, aCurIdx, aHeader);
        }

        /// <summary>
        /// Copy Constructor
        /// </summary>
        /// <param name=""></param>
        public ChatUserList(ChatUserList aItem)
        {
            Set(aItem);
        }

        #endregion

        #region Reset, Set

        /// <summary>
        /// Set fields to default values
        /// </summary>
        public void Reset()
        {
            mTheList = new List<ChatUserItem>();
            mCurIdx = -1;
            mHeader = "unused";
        }

        /// <summary>
        /// Set the values of every field according to arguments
        /// </summary>
        /// <param name=""></param>
        public void Set(List<ChatUserItem> aTheList, int aCurIdx, string aHeader)
        {
            mTheList = new List<ChatUserItem>(aTheList);
            mCurIdx = aCurIdx;
            mHeader = aHeader;
        }

        /// <summary>
        /// Copy our values from given ChatUserList
        /// </summary>
        /// <param name="aFrom"></param>
        public void Set(ChatUserList aFrom)
        {
            mTheList = new List<ChatUserItem>(aFrom.mTheList);
            mCurIdx = aFrom.mCurIdx;
            mHeader = aFrom.mHeader;
        }

        #endregion

        #region Standard Methods

        /// <summary>
        /// Add the given item to the end of the list.
        /// </summary>
        /// <param name="item"></param>
        public void AddItem(ChatUserItem item)
        {
            ChatUserItem tempItem = new ChatUserItem(item);
            mTheList.Add(tempItem);
        }

        /// <summary>
        /// Return the number of items in the list.
        /// </summary>
        /// <returns></returns>
        public int GetCount()
        {
            return mTheList.Count;
        }

        /// <summary>
        /// Return the item at the requested index.
        /// </summary>
        /// <param name="idx"></param>
        /// <returns></returns>
        public ChatUserItem GetItem(int idx)
        {
            if (idx >= mTheList.Count)
                return null;

            return mTheList[idx];
        }

        /// <summary>
        /// Remove the item at the given index
        /// </summary>
        /// <param name="aIdx"></param>
        /// <returns></returns>
        public bool RemoveAt(int aIdx)
        {
            if (aIdx < 0 || aIdx >= mTheList.Count)
                return false;

            mTheList.RemoveAt(aIdx);
            return true;
        }

        /// <summary>
        /// Allows foreach to work with our list
        /// </summary>
        /// <returns></returns>
        public IEnumerator GetEnumerator()
        {
            foreach (object item in mTheList)
            {
                yield return item;
            }
        }

        #endregion

        #region Find / Get Methods

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aHostName"></param>
        /// <returns></returns>
        public int FindIdTag(string aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                ChatUserItem item = mTheList[i];

                if (item.IdTagValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aHostName"></param>
        /// <returns></returns>
        public ChatUserItem GetIdTag(string aValue)
        {
            foreach (ChatUserItem item in mTheList)
            {
                if (item.IdTagValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aClient"></param>
        /// <returns></returns>
        public int FindClient(TcpClient aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                ChatUserItem item = mTheList[i];

                if (item.ClientValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aClient"></param>
        /// <returns></returns>
        public ChatUserItem GetClient(TcpClient aValue)
        {
            foreach (ChatUserItem item in mTheList)
            {
                if (item.ClientValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Return item at the current index, or null if none
        /// </summary>
        /// <returns></returns>
        public ChatUserItem GetCurrentItem()
        {
            if (mCurIdx < 0 || mCurIdx >= mTheList.Count)
                return null;

            return mTheList[mCurIdx];
        }

        #endregion

        #region Logging

        /// <summary>
        /// Return ToString for this class, or "null"
        /// </summary>
        /// <param name="aValue"></param>
        /// <returns></returns>
        public static string ToStringP(ChatUserList aValue)
        {
            if (aValue == null)
                return "null";
            else
                return aValue.ToString();
        }

        /// <summary>
        /// Return fields of this object as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "mTheList='"+mTheList.ToString()+"'"+", mCurIdx="+mCurIdx+", mHeader='"+mHeader+"'";
        }

        /// <summary>
        /// Log this object to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void Logit(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log this object to nlog - short version
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void LogitL(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitLToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log the contents of this class to String
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            sb.AppendLine();
            sb.AppendLine(aIndent+aHeader + " ChatUserList. CurIdx=" + mCurIdx + ", item count=" + mTheList.Count + ":" );

            int i;

            for (i = 0; i < mTheList.Count; i++)
                sb.AppendLine(mTheList[i].LogitToString(aLevel, aIndent + "    ", "[" + i + "]:"));

            sb.AppendLine(aIndent+"    "+"mHeader='"+mHeader+"'");

            sb.Append(aIndent + "End of ChatUserList");
            return sb.ToString();
        }

        /// <summary>
        /// Log the contents of this class to String - short version
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitLToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            sb.AppendLine();
            sb.AppendLine(aIndent+aHeader + " ChatUserList. CurIdx=" + mCurIdx + ", item count=" + mTheList.Count + ":" );

            int i;

            for (i = 0; i < mTheList.Count; i++)
                sb.AppendLine(mTheList[i].LogitLToString(aLevel, aIndent + "    ", "[" + i + "]:"));

            sb.AppendLine(aIndent+"    "+"mHeader='"+mHeader+"'");

            sb.Append(aIndent + "End of ChatUserList");
            return sb.ToString();
        }

        #endregion

        #region I/O Routines

        /// <summary>
        /// Convert our class to an XML string via Serialize. Can throw exceptions
        /// </summary>
        /// <returns></returns>
        public string ToXML()
        {
            try
            {
                // Force UTF-8
                XmlSerializer serializer = new XmlSerializer(typeof(ChatUserList));
                using (StringWriter writer = new IdLibrary.Utf8StringWriter())
                {
                    serializer.Serialize(writer, this);
                    return writer.ToString();
                }
            }
            catch (Exception ex)
            {
                logger.Error("ToXML: Exception '" + ex + "'");
                throw new Exception("ToXML", ex);
            }
        }

        /// <summary>
        /// Convert given xml string to our class. Throw exception on fail.
        /// </summary>
        /// <param name="xml"></param>
        /// <returns></returns>
        public ChatUserList FromXML(string xml)
        {
            ChatUserList ei = null;
            if ((xml != null) && (xml.Length > 0))
            {
                try
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(ChatUserList));
                    ei = (ChatUserList)serializer.Deserialize(new StringReader(xml));
                }
                catch (Exception ex)
                {
                    logger.Error("FromXML: Got exception '" + ex + "'");
                    throw new Exception("FromXML", ex);
                }
            }
            return ei;
        }

        /// <summary>
        /// Write out our class to given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool WriteFile(string fileName)
        {
            try
            {
                StreamWriter writer = new StreamWriter(fileName);
                writer.Write(this.ToXML());
                writer.Close();
            }
            catch (Exception ex)
            {
                logger.Info("WriteFile: Got exception '" + ex + "'");
                return false;
            }
            // Success
            return true;
        }

        /// <summary>
        /// Read our class from given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool ReadFile(string fileName)
        {
            try
            {
                StreamReader reader = new StreamReader(fileName);
                string contents = reader.ReadToEnd();
                reader.Close();

                // Set all our values from what we got back
                ChatUserList fi = FromXML(contents);
                this.Set(fi);
            }
            catch (Exception ex)
            {
                logger.Info("ReadFile: Got exception '" + ex + "'");
                return false;
            }

            // Success
            return true;
        }

        #endregion

        #region Tester

        /// <summary>
        /// Basic test routine item and list.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool Tester()
        {
            string TEST_FILE = "testChatUser.xml";
            ChatUserItem item1, item2, item3;
            ChatUserList list1, list2;
            int findIdx;
            bool result;

            // Test creating the items and building a list
            try
            {
                logger.Info("");

                // Create list #1
                list1 = new ChatUserList(new List<ChatUserItem>(), 2, "bbb");
                item1 = ChatUserItem.BuildSample(1);
                list1.AddItem(item1);
                logger.Info("Tester:2: 1 item list string=[" + list1.ToString() + "]");

                // Create list #2
                list2 = new ChatUserList(list1);

                foreach (ChatUserItem updateItem in list2)
                    updateItem.IdTagValue += "_1";

                item2 = ChatUserItem.BuildSample(2);
                list2.AddItem(item2);

                item3 = ChatUserItem.BuildSample(3);
                list2.AddItem(item3);

                list2.Logit(LogLevel.Info, "", "Tester:3: List2:");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Build list section got ex: " + ex);
                return false;
            }

            // Test logging and finds
            try
            {
                list2.Logit(LogLevel.Info, "", "Tester:4: created list:");

                // Supposed to fail
                findIdx = list2.FindIdTag("aaa");

                if (findIdx >= 0)
                {
                    logger.Info("Tester: FindIdTag(aaa) succeeded, returning false");
                    return false;
                }

                // Supposed to succeed
                findIdx = list2.FindIdTag(list2[0].IdTagValue);

                if (findIdx < 0)
                {
                    logger.Info("Tester:5: FindUpperID(list2[0].IdTagValue) failed, returning false");
                    return false;
                }

                // Test removal
                result = list2.RemoveAt(findIdx);

                if (result != true)
                {
                    logger.Info("Tester:6: list2.RemoveAt(" + findIdx + ") failed, returning false");
                    return false;
                }

                list2.LogitL(LogLevel.Info, "", "Tester: After remove, list2:");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Build list section got ex: " + ex);
                return false;
            }

            try
            {
                // Write list2 out to XML
                if (list2.WriteFile(TEST_FILE) == false)
                {
                    logger.Info("Tester:7: list2 WriteFile("+TEST_FILE+") failed, returning false");
                    return false;
                }

                // Read in as ul1
                result = list1.ReadFile(TEST_FILE);

                if (result != true)
                {
                    logger.Info("Tester:8: ReadFile(" + TEST_FILE + ") failed, returning false");
                    return false;
                }

                list1.Logit(LogLevel.Info, "", "Tester:9: Read "+TEST_FILE+":");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Read/Write list section got ex: " + ex);
                return false;
            }

            // Success
            return true;
        }

        /// <summary>
        /// Return a sample record ChatUserList.
        /// Return sample on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static ChatUserList BuildSample(string aUpperID, int aValue)
        {
            ChatUserItem item;
            ChatUserList rec;

            rec = new ChatUserList(new List<ChatUserItem>(), 1, "aaa");

            // Build list of 3 of same kind
            if (aValue == 1 || aValue == 2 || aValue == 3)
            {
                item = ChatUserItem.BuildSample(aValue);
                rec.AddItem(item);
                item = ChatUserItem.BuildSample(aValue);
                rec.AddItem(item);
                item = ChatUserItem.BuildSample(aValue);
                rec.AddItem(item);
            }
            else
            {
                item = ChatUserItem.BuildSample(-1);
                rec.AddItem(item);
                item = ChatUserItem.BuildSample(-1);
                rec.AddItem(item);
                item = ChatUserItem.BuildSample(-1);
                rec.AddItem(item);
            }

            return rec;
        }

        /// <summary>
        /// Drill down through sample record until non-class 2nd item and update it
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public bool UpdateSample()
        {
            logger.Trace("UpdateSample:1:");

            foreach (ChatUserItem rec in mTheList)
            {
                if (rec.UpdateSample() == false)
                    return false;
            }

            return true;
        }

        #endregion

        #region Class-Specific

        /// <summary>
        /// Return index of first entry that is marked invalid, or -1 if none
        /// </summary>
        /// <returns></returns>
        public int GetInvalidIndex()
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
                if (mTheList[i].IsValidValue == false)
                    return i;

            return -1;
        }

        /// <summary>
        /// Delete all items marked as invalid
        /// </summary>
        public void DeleteInvalidItems()
        {
            int idx;

            while ((idx = GetInvalidIndex()) != -1)
            {
                RemoveAt(idx);
            }
        }

        #endregion

        #endregion

        #region Private Access

        #region Private Members

        private List<ChatUserItem> mTheList;
        private int mCurIdx;
        private string mHeader;

        #endregion

        #endregion
    }
}
