﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Net.Sockets;
using System.Net;

namespace ChatSystem
{
    public class ListenWorker
    {
        TcpClient clientSocket;
        public bool mIsAlive = false;

        // This method will be called when the thread is started. 
        public void DoWork()
        {
            try
            {
                // First we start up listener
                Console.WriteLine("DoWork: Starting up listener");
                mIsAlive = true;
                TcpListener listener = new TcpListener(new IPEndPoint(IPAddress.Any, GeneralRoutines.DEFAULT_SERVER_PORT));
                listener.Start();

                // Check for stops between requests
                while (!_shouldStop)
                {
                    Console.WriteLine("DoWork: Calling AcceptTcpClient...");
                    clientSocket = listener.AcceptTcpClient();
                    string hostname = clientSocket.Client.RemoteEndPoint.ToString();

                    ChatUserItem user = new ChatUserItem(hostname+"-user", clientSocket, true);

                    // Hands off while we are updating
                    lock (Utility.mChatUsers)
                    {
                        Utility.mChatUsers.AddItem(user);
                    }
                }

                listener.Stop();
                mIsAlive = false;
                Console.WriteLine("DoWork: terminating gracefully.");
            }
            catch (Exception ex)
            {
                _shouldStop = true;
                Console.WriteLine("DoWork: Got ex: " + ex);
            }
        }

        public void RequestStop()
        {
            Console.WriteLine("RequestStop: started");
            _shouldStop = true;
        }

        public void Abort()
        {
            Console.WriteLine("Abort: started");
            mIsAlive = false;
            this.Abort();
        }

        // Volatile is used as hint to the compiler that this data 
        // member will be accessed by multiple threads. 
        private volatile bool _shouldStop;
    }
}
